#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h> 
#include <assert.h>
#include <sys/mman.h>
#include <limits.h>		/* for PAGESIZE */
#include <iostream.h>

/*
* a simple attempt to mprotect memory and catch the seg faults
*/

// some definitions
#define dprintf(x) if (DEBUG) printf x
#define usage(x) printf("Usage: %s [-s pagesToAlloc] [-d]\n", x)
#define DEFAULT_ALLOC_SIZE 10002 * getpagesize()
#ifndef PAGESIZE
#define PAGESIZE 4096
#endif

// externs
extern int errno;

// functions
void readArgs(int argc, char ** argv);
void protect(char * startaddr, int size, int prot);
char * getpagestart (char * ); 
long time_diff( struct timeval start, struct timeval end );
void seg_handler(int, siginfo_t *, void *); 

int allocSize = -1;	// size to allocate
bool DEBUG = false;
char * array;
int sigsCaught = 0;

int 
main (int argc, char ** argv) {
	readArgs(argc, argv);

	struct sigaction sa;
	sa.sa_sigaction = seg_handler;
	//sigfillset( &sa.sa_mask);
	sigemptyset( &sa.sa_mask);
	sa.sa_flags = SA_SIGINFO;

	if (sigaction (SIGSEGV, &sa, 0) != 0) {
		perror("can't sigaction");
		exit (-1);
	}

	array = (char *) malloc (allocSize);
	if ( ! array ) {
		perror("can't malloc");
		exit (-1);	
	}

	// protect each page twice, first for READ then change to WRITE 
	int pages = allocSize / PAGESIZE;
	struct timeval start, end;
	struct timezone tz;
	if ( gettimeofday( &start, &tz ) < 0 ) {
		perror( "Couldn't set time.\n" );
	}
	for (int j = 0; j < 2; j++) {
		for (int i = 1; i < pages - 1; i++) {
			protect( &(array[PAGESIZE * i]), PAGESIZE, PROT_READ );
			array[PAGESIZE * i]++; 
		}
	}
	if ( gettimeofday( &end, &tz ) < 0 ) {
		perror( "Couldn't set time.\n" );
	}

	long microseconds = time_diff( start, end ); 

	if (DEBUG) {
		cout << "Total time is " << microseconds << endl 
			<< "Sigscaught is " << sigsCaught << endl;
	}
	float eachOne = (float) microseconds / sigsCaught;
	cout << eachOne << endl;
	return 0;
}

#define MILLION 1000000

long
time_diff( struct timeval start, struct timeval end ) {

	long simplest = end.tv_sec * MILLION + end.tv_usec;
	simplest -= (start.tv_sec * MILLION + start.tv_usec);

	return simplest;
}

#ifdef linux

#include <asm/sigcontext.h>
#include <asm/ucontext.h>

char *
get_fault_address( void *context )
{
    return (char*)(((struct ucontext *)context)->uc_mcontext.cr2);
}


#endif


/* takes an address and returns the start boundary of the containing page */
char *
getpagestart (char * addr) {
	char * p;
	// align to a multiple of PAGESIZE, p will be >= to mem
	p = (char *)(((int) addr + PAGESIZE-1) & ~(PAGESIZE-1));
	if (p != addr) {
		assert (p > addr);
		p -= PAGESIZE;
	}	
	return p;
}

/* given a start addr and a size and a protection value, this
* will set those permissions for that memory
* NOTE: if the parameters are not page sized, this will extend the 
* protection to the page boundaries
*/
void
protect (char * startaddr, int size, int prot) {
	char * end = startaddr + size;
	char * p = getpagestart (startaddr);
	
	int diff = startaddr - p;
	size += diff; 
	if (size % PAGESIZE) {
		size += (PAGESIZE - (size % PAGESIZE));
		assert (size % PAGESIZE == 0);
	} 
	mprotect(p, size, prot);
	dprintf(("Protected %x, length %i, with prot of %i\n", p, size, prot));
}

void 
readArgs (int argc, char ** argv) {
	int curArg = 1;

	while (curArg < argc) {
		if ( ! strcmp(argv[curArg], "-s")) {
			allocSize = atoi(argv[++curArg]) * getpagesize();
		} else if ( ! strcmp(argv[curArg], "-d")) {
			DEBUG = true;
		} else {
			usage(argv[0]);
			exit(-1);
		}
		curArg++;
	}

	if (allocSize < 0) {
		allocSize = DEFAULT_ALLOC_SIZE;
	}
	dprintf(("allocSize is %i\n", allocSize));
}


/** 
DON"T do malloc or free in the signal handler !!!
*/
void
seg_handler(int signal, siginfo_t *info, void *context) {
	int saved_errno = errno;
	assert (signal == SIGSEGV);

    #ifdef linux
           info->si_addr = get_fault_address(context);
    #endif

	dprintf(("**caught a seg fault at %x, will reset prots.\n", info->si_addr));
	
	char * pagestart = getpagestart((char *)info->si_addr);

	protect (pagestart, PAGESIZE, PROT_READ | PROT_WRITE);
	errno = saved_errno;
	sigsCaught++;
} 
